from IPython.terminal.embed import InteractiveShellEmbed
import functools
import patoolib
import requests
import IPython
import inspect
import signal
import shutil
import psutil
import os
import sys

def to_underscore_style(text):
    """Changes "Something like this" to "something_like_this"."""
    text = text.lower().replace(" ", "_").replace("-", "_")
    return ''.join(x for x in text if x.isalpha() or x.isdigit() or x == "_")

def ipython(message=None, frame=None):
    """Launch into customized IPython with greedy autocompletion and no prompt to exit."""
    config = IPython.Config({
        'InteractiveShell': {'confirm_exit': False, },
        'IPCompleter': {'greedy': True, }
    })
    InteractiveShellEmbed.instance(config=config)(message, local_ns=frame.f_locals, global_ns=frame.f_globals)

def ipython_embed(message=None, functions_above=0):
    frame = inspect.stack()[functions_above + 1][0]
    ipython(message, frame)

def stop_ipython():
    from IPython.terminal.interactiveshell import TerminalInteractiveShell
    TerminalInteractiveShell.exit_now = True


def _write(handle, message):
    if os.isatty(handle.fileno()):
        handle.write(message)
    else:
        handle.write(bytes(message, 'utf8'))
    handle.flush()

def log(message):
    """Output to stdout."""
    import sys
    _write(sys.stdout, message)

def warn(message):
    """Output to stderr."""
    import sys
    _write(sys.stderr, message)


def do_exit(signal, frame):
    """Just exit."""
    sys.exit(1)

def signals_trigger_exit():
    """Make all signals cause a system exit."""
    signal.signal(signal.SIGINT, do_exit)
    signal.signal(signal.SIGTERM, do_exit)
    signal.signal(signal.SIGHUP, do_exit)
    signal.signal(signal.SIGQUIT, do_exit)

def ignore_signals():
    """Ignore all signals (e.g. ctrl-C)."""
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)
    signal.signal(signal.SIGHUP, signal.SIG_IGN)
    signal.signal(signal.SIGQUIT, signal.SIG_IGN)

def signal_pass_on_to_separate_process_group(pid):
    def pass_on_signal(pid, signum, frame):
        os.kill(pid, signum)

    signal.signal(signal.SIGINT, functools.partial(pass_on_signal, pid))
    signal.signal(signal.SIGTERM, functools.partial(pass_on_signal, pid))
    signal.signal(signal.SIGHUP, functools.partial(pass_on_signal, pid))
    signal.signal(signal.SIGQUIT, functools.partial(pass_on_signal, pid))

def extract_archive(filename, directory):
    patoolib.extract_archive(filename, outdir=directory)


def download_file(downloaded_file_name, url):
    TOTAL_BLOCKS = 79

    downloaded_file_name_temp = "{}.part".format(downloaded_file_name)

    if os.path.exists(downloaded_file_name_temp):
        os.remove(downloaded_file_name_temp)

    if not os.path.exists(downloaded_file_name):
        with open(downloaded_file_name_temp, "wb") as downloaded_file_handle:
            log("Downloading {}\n".format(url))
            response = requests.get(url, stream=True)
            total_length = response.headers.get('content-length')
            previous_done = None

            if total_length is None: # no content length header
                downloaded_file_handle.write(response.content)
            else:
                downloaded_bytes = 0
                total_length = int(total_length)
                for data in response.iter_content():
                    downloaded_bytes += len(data)
                    downloaded_file_handle.write(data)
                    blocks_done = int(TOTAL_BLOCKS * float(downloaded_bytes) / float(total_length))
                    percentage_done = "{:.1f}".format(
                        100 * float(downloaded_bytes) / float(total_length)
                    )
                    if percentage_done != previous_done:
                        previous_done = percentage_done
                        log("\r{}% [{}{}]".format(
                            percentage_done,
                            '=' * blocks_done,
                            ' ' * (TOTAL_BLOCKS - blocks_done))
                        )

        log("\n")
        shutil.move(downloaded_file_name_temp, downloaded_file_name)
    else:
        log("Downloaded : {} already found\n".format(downloaded_file_name))




from hitchtest.executionengine import ExecutionEngine
from hitchtest.arguments import Arguments
from hitchtest.step import Step
from hitchtest.scenario import Scenario
from hitchtest.test import Test
from hitchtest.module import Module
from hitchtest.utils import to_underscore_style, ipython, ipython_embed
from hitchtest import commandline
from hitchtest.package import HitchPackage
from os import path


HITCHTEST_DIR = path.dirname(path.realpath(__file__))



import signal


class SignalManager(object):
    NAMES = dict((getattr(signal, n), n) \
        for n in dir(signal) if n.startswith('SIG') and '_' not in n )

    def __init__(self, abort_handler):
        self._original_signals = {}
        for sig in [signal.SIGTERM, signal.SIGHUP, signal.SIGQUIT, signal.SIGINT]:
            self._save_original_signal_handler(sig)
            self.attach_handler(sig, abort_handler)

    def _save_original_signal_handler(self, sig):
        self._original_signals[sig] = signal.getsignal(sig)

    def turn_off_signal_handler(self, sig):
        if sig in self._original_signals:
            signal.signal(sig, self._original_signals[sig])
        else:
            raise RuntimeError("Signal {} not in original signals.".format(sig))

    def attach_handler(self, sig, handler):
        signal.signal(sig, handler)



from hitchtest.environment.utils import version_comparison, return_code_zero, check_output_lines
from collections import Counter
from os import path
import struct
import psutil
import sys
import re


class HitchEnvironmentException(Exception):
    pass

def freeports(required_ports):
    """Verifies that none of the required_ports are not in use. Raise exception if they are."""
    unusable_ports = []
    import socket
    for port in required_ports:
        socket_to_try = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_to_try.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            socket_to_try.bind(('', port))
            socket_to_try.listen(1)
        except socket.error as e:
            unusable_ports.append(str(port))
        finally:
            socket_to_try.close()

    if len(unusable_ports) > 0:
        raise HitchEnvironmentException(
            "Required network port(s): {} not usable.".format(', '.join(unusable_ports))
        )

def debs(packages):
    """Verify that a list of debs are installed. Ignore if not a deb based system."""
    if return_code_zero(["which", "dpkg"]):
        if not return_code_zero(["dpkg", "--list",] + packages):
            raise HitchEnvironmentException(
                "sudo apt-get install {} : required for test to run".format(' '.format(packages))
            )

def brew(packages):
    """Verify that a list of brew packages are installed. Ignore if not a mac."""
    if sys.platform == "darwin":
        version_list = [x for x in check_output_lines(["brew", "list", "--versions"] + packages) if x != ""]
        if len(version_list) < len(packages):
            raise HitchEnvironmentException(
                "brew install {} : required for test to run".format(' '.join(packages))
            )

def systembits(bits):
    """Verify that a system is 64 or 32 bit."""
    if not bits == struct.calcsize("P") * 8:
        raise HitchEnvironmentException(
            "{} bit system required to run test. This system is {} bit".format(str(bits))
        )

def internet_detected_after(timeout):
    """Verify that a system is connected to the internet."""
    if not return_code_zero(["ping", "-c", "1", "-W", str(timeout), "8.8.8.8"]):
        raise HitchEnvironmentException(
            "No internet detected after {} seconds. Ping failed.".format(timeout)
        )

def approved_platforms(platforms):
    """Verify that the test is running on an approved platform."""
    if not sys.platform in platforms:
        raise HitchEnvironmentException(
            "This platform is {}. This test will only run on {}.".format(' '.format(platforms))
        )


# TODO : rpm package availability.
# TODO : arch package availability.
# TODO : all other package managers
# TODO : Kernel SHMMAX, glibc version, CPU usage,
# TODO : psutil.phymem_usage().available / 1024 / 1024
# TODO : psutil.cpu_percent()
# TODO : psutil.cpu_count()
# TODO : Linux Kernel version
# TODO : Mac OS X kernel version
# TODO : ulimit
# TODO : nslookup -timeout=1 microsoft.com


#def linux_distro(distros):
    #if path.exists("/etc/issue"):
        #with open("/etc/issue", "r") as issue_handle:
            #issue = issue_handle.read()

        #if "ubuntu" in distros and issue.startswith("Ubuntu"):
            #return True
        #if "redhat" in distros and issue.startswith("Red Hat"):
            #return True
        #return False
    #return True

#def ubuntu_version(version_check):
    #if path.exists("/etc/issue"):
        #with open("/etc/issue", "r") as issue_handle:
            #issue = issue_handle.read()
        #version = re.search("[0-9]+\.[0-9]+\.?[0-9]*", issue).group(0)
        #return version_comparison(version_check, version)
    #return True



from hitchtest.arguments import Arguments
from hitchtest import utils


class Step(object):
    def __init__(self, yaml_step, index):
        self.index = int(index)
        if type(yaml_step) is str:
            self.name = str(yaml_step)
            self.arguments = Arguments(None)
        elif type(yaml_step) is dict and len(yaml_step.keys()) == 1:
            self.name = list(yaml_step.keys())[0]
            self.arguments = Arguments(list(yaml_step.values())[0])
        else:
            raise RuntimeError("Invalid YAML in step '{}'".format(yaml_step))

    def __str__(self):
        return self.name

    def underscore_case_name(self):
        return utils.to_underscore_style(self.name)

    def to_dict(self):
        return {'index': self.index, 'name': self.name, 'arguments': self.arguments.to_dict(), }

    def run(self, engine):
        if self.arguments.is_none:
            getattr(engine, self.underscore_case_name())()
        elif self.arguments.single_argument:
            getattr(engine, self.underscore_case_name())(self.arguments.argument)
        else:
            getattr(engine, self.underscore_case_name())(**self.arguments.pythonized_kwargs())
